@@ -1,12 +1,9 @@
 #!/usr/bin/env python
+# -*- coding: utf-8 -*-
 
-##
-## TODO:
-##
-##  - use an external template to format changelog
-##
-
+from __future__ import print_function
 
+import locale
 import re
 import os
 import os.path
@@ -15,6 +12,7 @@ import glob
 import textwrap
 import datetime
 import collections
+
 from subprocess import Popen, PIPE
 
 try:
@@ -27,11 +25,10 @@ try:
 except ImportError:
     mako = None
 
+PY3 = sys.version_info[0] >= 3
 
-
-usage_msg = """usage: %(exname)s [REPOS]
-
-Run this command in a git repository to get a ReST changelog in stdout.
+usage_msg = """usage: %(exname)s"""
+help_msg = """Run this command in a git repository to output a formatted changelog
 
 %(exname)s uses a config file to filter meaningfull commit or do some
  formatting in commit messages thanks to a config file.
@@ -39,16 +36,22 @@ Run this command in a git repository to get a ReST changelog in stdout.
 Config file location will be resolved in this order:
 
   - in shell environment variable GITCHANGELOG_CONFIG_FILENAME
-  - in git configuration: ``git config gitchangelog.rc-patch``
+  - in git configuration: ``git config gitchangelog.rc-path``
   - as '.%(exname)s.rc' in the root of the current git repository
-  - as '~/.%(exname)s.rc'
-  - as '/etc/%(exname)s'
 
 """
 
+full_help_msg = "%s\n\n%s" % (usage_msg, help_msg)
+
 
 class ShellError(Exception):
-    pass
+
+    def __init__(self, msg, errlvl=None, command=None, out=None, err=None):
+        self.errlvl = errlvl
+        self.command = command
+        self.out = out
+        self.err = err
+        super(ShellError, self).__init__(msg)
 
 
 def die(msg=None):
@@ -69,93 +72,28 @@ def available_in_config(f):
     return f
 
 
-def load_config_file(filename, fail_if_not_present=True):
+def load_config_file(filename, default_filename=None,
+                     fail_if_not_present=True):
     """Loads data from a config file."""
 
     config = _config_env.copy()
-
-    if os.path.exists(filename):
-        try:
-            execfile(filename, config)
-        except SyntaxError, e:
-            die('Syntax error in config file: %s\n'
-                'Line %i offset %i\n' % (filename, e.lineno, e.offset))
-    else:
-        if fail_if_not_present:
-            die('%r config file is not found and is required.' % (filename, ))
+    for fname in [default_filename, filename]:
+        if fname and os.path.exists(fname):
+            try:
+                with open(fname) as f:
+                    code = compile(f.read(), fname, 'exec')
+                    exec(code, config)
+            except SyntaxError as e:
+                die('Syntax error in config file: %s\n'
+                    'Line %i offset %i\n'
+                    % (fname, e.lineno, e.offset))
+        else:
+            if fail_if_not_present:
+                die('%r config file is not found and is required.' % (filename, ))
 
     return config
 
 
-##
-## Common functions
-##
-
-def inflate_dict(dct, sep=".", deep=-1):
-    """Inflates a flattened dict.
-
-    Will look in simple dict of string key with string values to
-    create a dict containing sub dicts as values.
-
-    Samples are better than explanation:
-
-        >>> from pprint import pprint as pp
-        >>> pp(inflate_dict({'a.x': 3, 'a.y': 2}))
-        {'a': {'x': 3, 'y': 2}}
-
-    The keyword argument ``sep`` allows to change the separator used
-    to get subpart of keys:
-
-        >>> pp(inflate_dict({'etc/group': 'geek', 'etc/user': 'bob'}, "/"))
-        {'etc': {'group': 'geek', 'user': 'bob'}}
-
-    Warning: you cannot associate a value to a section:
-
-        >>> inflate_dict({'section.key': 3, 'section': 'bad'})
-        Traceback (most recent call last):
-        ...
-        TypeError: 'str' object does not support item assignment
-
-    Of course, dict containing only keys that doesn't use separator will be
-    returned without changes:
-
-        >>> inflate_dict({})
-        {}
-        >>> inflate_dict({'a': 1})
-        {'a': 1}
-
-    Argument ``deep``, is the level of deepness allowed to inflate dict:
-
-        >>> pp(inflate_dict({'a.b.c': 3, 'a.d': 4}, deep=1))
-        {'a': {'b.c': 3, 'd': 4}}
-
-    Of course, a deepness of 0 won't do anychanges, whereas deepness of -1 is
-    the default value and means infinite deepness:
-
-        >>> pp(inflate_dict({'a.b.c': 3, 'a.d': 4}, deep=0))
-        {'a.b.c': 3, 'a.d': 4}
-
-    """
-
-    def mset(dct, k, v, sep=".", deep=-1):
-        if deep == 0 or sep not in k:
-            dct[k] = v
-        else:
-            khead, ktail = k.split(sep, 1)
-            if khead not in dct:
-                dct[khead] = {}
-            mset(dct[khead], ktail, v,
-                 sep=sep,
-                 deep=-1 if deep < 0 else deep - 1)
-
-    res = {}
-    ## sorting keys ensures that colliding values if any will be string
-    ## first set first so mset will crash with a TypeError Exception.
-    for k in sorted(dct.keys()):
-        mset(res, k, dct[k], sep, deep)
-    return res
-
-
 ##
 ## Text functions
 ##
@@ -178,12 +116,12 @@ def indent(text, chars="  ", first=None):
 
     >>> string = 'This is first line.\\nThis is second line\\n'
 
-    >>> print indent(string, chars="| ") # doctest: +NORMALIZE_WHITESPACE
+    >>> print(indent(string, chars="| "))  # doctest: +NORMALIZE_WHITESPACE
     | This is first line.
     | This is second line
     |
 
-    >>> print indent(string, first="- ") # doctest: +NORMALIZE_WHITESPACE
+    >>> print(indent(string, first="- "))  # doctest: +NORMALIZE_WHITESPACE
     - This is first line.
       This is second line
 
@@ -204,7 +142,7 @@ def paragraph_wrap(text, regexp="\n\n"):
     >>> string = 'This is first paragraph which is quite long don\'t you \
     ... think ? Well, I think so.\n\nThis is second paragraph\n'
 
-    >>> print paragraph_wrap(string) # doctest: +NORMALIZE_WHITESPACE
+    >>> print(paragraph_wrap(string)) # doctest: +NORMALIZE_WHITESPACE
     This is first paragraph which is quite long don't you think ? Well, I
     think so.
     This is second paragraph
@@ -222,31 +160,34 @@ def paragraph_wrap(text, regexp="\n\n"):
 ##
 
 
-def cmd(command):
 
+def cmd(command, env=None):
     p = Popen(command, shell=True,
               stdin=PIPE, stdout=PIPE, stderr=PIPE,
-              close_fds=True)
+              close_fds=True, env=env,
+              universal_newlines=False)
     stdout, stderr = p.communicate()
-    return stdout, stderr, p.returncode
+    return (stdout.decode(locale.getpreferredencoding()),
+            stderr.decode(locale.getpreferredencoding()),
+            p.returncode)
 
 
-def wrap(command, quiet=True, ignore_errlvls=[0]):
+def wrap(command, ignore_errlvls=[0], env=None):
     """Wraps a shell command and casts an exception on unexpected errlvl
 
-    >>> wrap('/tmp/lsdjflkjf') # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
+    >>> wrap('/tmp/lsdjflkjf') # doctest: +ELLIPSIS +IGNORE_EXCEPTION_DETAIL
     Traceback (most recent call last):
     ...
     ShellError: Wrapped command '/tmp/lsdjflkjf' exited with errorlevel 127.
       stderr:
-      | /bin/sh: /tmp/lsdjflkjf: not found
+      | /bin/sh: .../tmp/lsdjflkjf: not found
 
-    >>> wrap('echo hello') # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
-    'hello\\n'
+    >>> print(wrap('echo hello'),  end='')
+    hello
 
     """
 
-    out, err, errlvl = cmd(command)
+    out, err, errlvl = cmd(command, env=env)
 
     if errlvl not in ignore_errlvls:
 
@@ -262,7 +203,8 @@ def wrap(command, quiet=True, ignore_errlvls=[0]):
         formatted = '\n'.join(formatted)
 
         raise ShellError("Wrapped command %r exited with errorlevel %d.\n%s"
-                        % (command, errlvl, indent(formatted, chars="  ")))
+                         % (command, errlvl, indent(formatted, chars="  ")),
+                         errlvl=errlvl, command=command, out=out, err=err)
     return out
 
 
@@ -286,36 +228,50 @@ class SubGitObjectMixin(object):
         return self._repos.swrap(*args, **kwargs)
 
 
+GIT_FORMAT_KEYS = {
+    'sha1': "%H",
+    'subject': "%s",
+    'author_name': "%an",
+    'author_date': "%ad",
+    'author_date_timestamp': "%at",
+    'committer_name': "%cn",
+    'committer_date_timestamp': "%ct",
+    'raw_body': "%B",
+    'body': "%b",
+}
+
+
 class GitCommit(SubGitObjectMixin):
 
     def __init__(self, identifier, repos):
         super(GitCommit, self).__init__(repos)
-
         self.identifier = identifier
 
-        if identifier is "LAST":
-            identifier = self.swrap("git log --format=%H | tail -n 1")
+    def __getattr__(self, label):
+        """Completes commits attributes upon request."""
+        attrs = GIT_FORMAT_KEYS.keys()
+        if label not in attrs:
+            return super(GitCommit, self).__getattr__(label)
 
-        attrs = {'sha1': "%h",
-                 'subject': "%s",
-                 'author_name': "%an",
-                 'author_date': "%ad",
-                 'author_date_timestamp': "%at",
-                 'committer_name': "%cn",
-                 'committer_date_timestamp': "%ct",
-                 'raw_body': "%B",
-                 'body': "%b",
-                 }
-        aformat = "%x00".join(attrs.values())
+        identifier = self.identifier
+        if identifier == "LAST":
+            identifier = self.swrap(
+                "git rev-list --first-parent --max-parents=0 HEAD")
+
+        ## Compute only missing information
+        missing_attrs = [l for l in attrs if not l in self.__dict__]
+        aformat = "%x00".join(GIT_FORMAT_KEYS[l]
+                              for l in missing_attrs)
         try:
-            ret = self.swrap("git show -s %r --pretty=format:%s"
-                       % (identifier, aformat))
+            ret = self.swrap("git show -s %s --pretty=format:%s"
+                             % (identifier, aformat))
         except ShellError:
-            raise ValueError("Given commit identifier %r doesn't exists"
-                             % identifier)
+            raise ValueError("Given commit identifier %s doesn't exists"
+                             % self.identifier)
         attr_values = ret.split("\x00")
-        for attr, value in zip(attrs.keys(), attr_values):
+        for attr, value in zip(missing_attrs, attr_values):
             setattr(self, attr, value.strip())
+        return getattr(self, label)
 
     @property
     def date(self):
@@ -331,19 +287,6 @@ class GitCommit(SubGitObjectMixin):
     def __hash__(self):
         return hash(self.sha1)
 
-    def __sub__(self, value):
-        if not isinstance(value, GitCommit):
-            raise TypeError("Invalid type for %r in operation" % value)
-        if self.sha1 == value.sha1:
-            return []
-        commits = self.swrap('git rev-list %s..%s'
-                             % (value.sha1, self.sha1))
-        if not commits:
-            raise ValueError('Seems that %r is earlier than %r'
-                             % (self.identifier, value.identifier))
-        return [GitCommit(commit, self._repos)
-                for commit in reversed(commits.split('\n'))]
-
     def __repr__(self):
         return "<%s %r>" % (self.__class__.__name__, self.identifier)
 
@@ -359,6 +302,101 @@ def normpath(path, cwd=None):
     return os.path.normpath(os.path.join(cwd, path))
 
 
+class GitConfig(SubGitObjectMixin):
+    """Interface to config values of git
+
+    Let's create a fake GitRepos:
+
+        >>> from minimock import Mock
+        >>> repos = Mock("gitRepos")
+
+    Initialization:
+
+        >>> cfg = GitConfig(repos)
+
+    Query, by attributes or items:
+
+        >>> repos.swrap.mock_returns = "bar"
+        >>> cfg.foo
+        Called gitRepos.swrap("git config 'foo'")
+        'bar'
+        >>> cfg["foo"]
+        Called gitRepos.swrap("git config 'foo'")
+        'bar'
+        >>> cfg.get("foo")
+        Called gitRepos.swrap("git config 'foo'")
+        'bar'
+        >>> cfg["foo.wiz"]
+        Called gitRepos.swrap("git config 'foo.wiz'")
+        'bar'
+
+    Notice that you can't use attribute search in subsection as ``cfg.foo.wiz``
+    That's because in git config files, you can have a value attached to
+    an element, and this element can also be a section.
+
+    Nevertheless, you can do:
+
+        >>> getattr(cfg, "foo.wiz")
+        Called gitRepos.swrap("git config 'foo.wiz'")
+        'bar'
+
+    Default values
+    --------------
+
+    get item, and getattr default values can be used:
+
+        >>> del repos.swrap.mock_returns
+        >>> repos.swrap.mock_raises = ShellError('Key not found',
+        ...                                      errlvl=1, out="", err="")
+
+        >>> getattr(cfg, "foo", "default")
+        Called gitRepos.swrap("git config 'foo'")
+        'default'
+
+        >>> cfg["foo"]  ## doctest: +ELLIPSIS
+        Traceback (most recent call last):
+        ...
+        KeyError: 'foo'
+
+        >>> getattr(cfg, "foo")  ## doctest: +ELLIPSIS
+        Traceback (most recent call last):
+        ...
+        AttributeError...
+
+        >>> cfg.get("foo", "default")
+        Called gitRepos.swrap("git config 'foo'")
+        'default'
+
+        >>> print("%r" % cfg.get("foo"))
+        Called gitRepos.swrap("git config 'foo'")
+        None
+
+    """
+
+    def __init__(self, repos):
+        super(GitConfig, self).__init__(repos)
+
+    def __getattr__(self, label):
+        cmd = "git config %r" % str(label)
+        try:
+            res = self.swrap(cmd)
+        except ShellError as e:
+            if e.errlvl == 1 and e.out == "" and e.err == "":
+                raise AttributeError("key %r is not found in git config."
+                                     % label)
+            raise
+        return res
+
+    def get(self, label, default=None):
+        return getattr(self, label, default)
+
+    def __getitem__(self, label):
+        try:
+            return getattr(self, label)
+        except AttributeError:
+            raise KeyError(label)
+
+
 class GitRepos(object):
 
     def __init__(self, path):
@@ -367,6 +405,21 @@ class GitRepos(object):
         ## will be done from this location.
         self._orig_path = os.path.abspath(path)
 
+        ## verify ``git`` command is accessible:
+        try:
+            self._git_version = self.swrap("git version")
+        except ShellError:
+            raise EnvironmentError(
+                "Required ``git`` command not found or broken in $PATH. "
+                "(calling ``git version`` failed.)")
+
+        ## verify that we are in a git repository
+        try:
+            self.swrap("git remote")
+        except ShellError:
+            raise EnvironmentError("Not in a git repository. "
+                "(calling ``git remote`` failed.)")
+
         self.bare = self.swrap("git rev-parse --is-bare-repository") == "true"
         self.toplevel = None if self.bare else \
                         self.swrap("git rev-parse --show-toplevel")
@@ -376,9 +429,7 @@ class GitRepos(object):
 
     @property
     def config(self):
-        all_options = self.swrap("git config -l")
-        dct_options = dict(l.split("=", 1) for l in all_options.split('\n'))
-        return inflate_dict(dct_options)
+        return GitConfig(self)
 
     def swrap(self, command, **kwargs):
         """Essential force the CWD of the command to be in self._orig_path"""
@@ -389,27 +440,31 @@ class GitRepos(object):
     @property
     def tags(self):
         tags = self.swrap('git tag -l').split("\n")
-        while '' in tags:
-            tags.remove('')
-        return sorted([GitCommit(tag, self) for tag in tags],
+        return sorted([GitCommit(tag, self) for tag in tags if tag != ''],
                       key=lambda x: int(x.committer_date_timestamp))
 
-    def __getitem__(self, key):
+    def log(self, start="HEAD"):
 
-        if isinstance(key, basestring):
-            return GitCommit(key, self)
+        aformat = "%x00".join(GIT_FORMAT_KEYS.values())
+        try:
+            ret = self.swrap(
+                "git log %r -z --first-parent --pretty=format:%s"
+                % (start, aformat))
+        except ShellError:
+            raise ValueError("Given commit identifier %r doesn't exists"
+                             % start)
 
-        if isinstance(key, slice):
-            start, stop = key.start, key.stop
+        def mk_commit(dct):
+            """Creates an already set commit from a dct"""
+            c = GitCommit(dct["sha1"], self)
+            for k, v in dct.items():
+                setattr(c, k, v)
+            return c
 
-            if start is None:
-                start = GitCommit('LAST', self)
-
-            if stop is None:
-                stop = GitCommit('HEAD', self)
-
-            return stop - start
-        raise NotImplementedError("Unsupported getitem %r object." % key)
+        values = iter(ret.split('\x00'))
+        while True: ## values.next() will eventualy raise a StopIteration
+            yield mk_commit(dict([(key, next(values))
+                                  for key in GIT_FORMAT_KEYS]))
 
 
 def first_matching(section_regexps, string):
@@ -547,10 +602,10 @@ if mako:
         template_path = os.path.join(template_dir, "%s.tpl" % template_name)
 
         if not os.path.exists(template_path):
-            print "Available mako templates:\n - " + \
+            print("Available mako templates:\n - " +
                   " - ".join(os.path.basename(f).split(".")[-1]
                              for f in glob.glob(os.path.join(template_dir,
-                                                             "*.tpl")))
+                                                             "*.tpl"))))
             die("No %r a valid mako template name." % template_name)
 
         template = mako.template.Template(filename=template_path)
@@ -602,7 +657,7 @@ def changelog(repository,
 
     """
 
-    def new_version(tag, date, opts) :
+    def new_version(tag, date, opts):
         title = "%s (%s)" % (tag, date) if tag else \
                 opts["unreleased_version_label"]
         return {"label": title,
@@ -633,7 +688,7 @@ def changelog(repository,
 
     section_order = [k for k, _v in section_regexps]
 
-    for commit in reversed(repository[:]):
+    for commit in repository.log():
 
         tags_of_commit = [tag for tag in tags
                           if tag == commit]
@@ -659,7 +714,7 @@ def changelog(repository,
         ## Replace content in commit subject
 
         subject = commit.subject
-        for regexp, replacement in replace_regexps.iteritems():
+        for regexp, replacement in replace_regexps.items():
             subject = re.sub(regexp, replacement, subject)
 
         ## Finally storing the commit in the matching section
@@ -669,7 +724,7 @@ def changelog(repository,
         sections[matched_section].append({
             "author": commit.author_name,
             "subject": subject,
-            "body": commit.body
+            "body": commit.body,
             })
 
     ## Don't forget last commits:
@@ -680,32 +735,55 @@ def changelog(repository,
 
     return output_engine(data=changelog, opts=opts)
 
-
 ##
 ## Main
 ##
 
-
 def main():
 
+    reference_config = os.path.join(
+            os.path.dirname(os.path.realpath(__file__)),
+            "gitchangelog.rc.reference")
+
     basename = os.path.basename(sys.argv[0])
     if basename.endswith(".py"):
         basename = basename[:-3]
 
-    if len(sys.argv) == 1:
-        repos = "."
-    elif len(sys.argv) == 2:
-        if sys.argv[1] == "--help":
-            print usage_msg % {'exname': basename}
-            sys.exit(0)
-        repos = sys.argv[1]
-    else:
-        die('usage: %s [REPOS]\n' % basename)
+    if "-h" in sys.argv or "--help" in sys.argv:
+        print(full_help_msg % {'exname': basename})
+        sys.exit(0)
 
-    ## warning: not safe (repos is given by the user)
-    repository = GitRepos(repos)
+    if len(sys.argv) > 1 and sys.argv[1] != "init":
+        die(usage_msg % {'exname': basename})
+
+    try:
+        repository = GitRepos(".")
+    except EnvironmentError as e:
+        die(e.message)
+
+    repository_config = '%s/.%s.rc' % (repository.toplevel, basename) \
+                        if not repository.bare else None
+
+    if len(sys.argv) == 2 and sys.argv[1] == "init":
+        import shutil
+        if repository_config is None:
+            die("``init`` of bare repository not supported.")
+        if os.path.exists(repository_config):
+            die("File %r already exists." % repository_config)
+        shutil.copyfile(reference_config,
+                        repository_config)
+        print("File %r created.")
+        sys.exit(0)
+
+    try:
+        gc_rc = repository.config.get("gitchangelog.rc-path")
+    except ShellError as e:
+        sys.stderr.write(
+            "Error parsing git config: %s."
+            " Won't be able to read 'rc-path' if defined.\n"
+            % (str(e)))
+        gc_rc = None
 
-    gc_rc = repository.config.get("gitchangelog", {}).get('rc-path')
     gc_rc = normpath(gc_rc, cwd=repository.toplevel) if gc_rc else None
 
     ## config file lookup resolution
@@ -714,8 +792,9 @@ def main():
         (True, lambda: gc_rc),
         (False, lambda: ('%s/.%s.rc' % (repository.toplevel, basename)) \
                  if not repository.bare else None),
-        (False, lambda: os.path.expanduser('~/.%s.rc' % basename)),
-        (False, lambda: '/etc/%s.rc' % basename),
+        ## Removed to enforce per-repository gitchangelog file.
+        # (False, lambda: os.path.expanduser('~/.%s.rc' % basename)),
+        # (False, lambda: '/etc/%s.rc' % basename),
         ]:
         changelogrc = fun()
         if changelogrc:
@@ -727,12 +806,12 @@ def main():
             else:
                 break
 
-    if not changelogrc or not os.path.exists(changelogrc):
-        die("No %s config file found anywhere." % basename)
+    config = load_config_file(
+        os.path.expanduser(changelogrc),
+        default_filename=reference_config,
+        fail_if_not_present=False)
 
-    config = load_config_file(os.path.expanduser(changelogrc))
-
-    print changelog(repository,
+    content = changelog(repository,
         ignore_regexps=config['ignore_regexps'],
         replace_regexps=config['replace_regexps'],
         section_regexps=config['section_regexps'],
@@ -742,11 +821,15 @@ def main():
         output_engine=config.get("output_engine", rest_py),
     )
 
+    if PY3:
+        print(content)
+    else:
+        print(content.encode(locale.getpreferredencoding()))
+
 ##
 ## Launch program
 ##
 
-
 if __name__ == "__main__":
 
     # import doctest
